// Control fan speed with MEI

#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>

#include <unistd.h>
#include <sys/signal.h>

#include "QstCfg.h"
#include "QstComm.h"
#include "AccessQst.h"

#include "defines.h"

char *sensorname[MAXSENSORS];

fan_t fan[MAXFANS];

int interval=0;

extern void detect(void);
extern int readconfig(char *name);

// Read temperature
float readtemp(char *dev)
{
	float val;

#ifdef DEBUG
	printf("Reading temperature from: %s\n",dev);
#endif

	FILE *f=fopen(dev,"r");

	if(f==NULL)
		return -1;

	fscanf(f,"%f",&val);
	fclose(f);

#ifdef DEBUG
	printf("Got: %d\n",(int)val);
#endif

	return val/1000.0;
}

// Main fan/temperature logic
void fanlogic(void)
{
	FILE *f;

	float duty,temp,tmp;

	int pn,i,changed,heartbeat=0;

	while(1)
	{
		// Check all core temps and find the highest
		temp=-1000;
		for(i=0;sensorname[i];i++)
		{
			tmp=readtemp(sensorname[i]);
			if(tmp>temp)
				temp=tmp;
		}

		// Calculate duty cycle for each dynamic fan
		for(i=0,changed=0;i<MAXFANS;i++)
		{
			if(fan[i].type==DYNAMIC)
			{
				if(temp<fan[i].mintemp) // Comfortably cool
					fan[i].duty=fan[i].minduty;
				else
				{
					if(temp>fan[i].maxtemp) // Too hot
						fan[i].duty=fan[i].maxduty;
					else // In-between
					{
						fan[i].duty=(temp-fan[i].mintemp)/
								    (fan[i].maxtemp-fan[i].mintemp); // 0..1
						fan[i].duty=fan[i].minduty+
									fan[i].duty*(fan[i].maxduty-fan[i].minduty);
					}
				}
			}

			if(fan[i].oldduty!=fan[i].duty) // Minimize changes to avoid trouble
				changed++;

			fan[i].oldduty=fan[i].duty;
		}

#ifdef DEBUG
		puts("Writing heartbeat.");
#endif
		f=fopen(HEARTBEAT_FILE,"w");
		if(f==NULL)
		{
			printf("*** Cannot open %s for writing.\n",HEARTBEAT_FILE);
			exit(EXIT_FAILURE);
		}

		fprintf(f,"%d",heartbeat++),
		fclose(f);

#ifdef DEBUG
		puts("Starting to sleep.");
#endif
		sleep(interval);

		if(!changed)
			continue;

		if(!InitializeQst())
		{
#ifdef DEBUG
			puts("*** MEI initialization problem.");
#endif
			exit(EXIT_FAILURE);
		}

		for(i=0;i<MAXFANS;i++)
		{
			if(fan[i].type!=UNUSED)
			{
#ifdef DEBUG
				printf("Setting fan %d duty cycle: %f\n",i,fan[i].duty);
#endif
				SetDutyManualQst(i,fan[i].duty);
			}
		}

		CleanupQst();
	}
}

void handler(int sig)
{
	int i;

	if(sig==SIGTERM)
	{
		puts("Stopping meifand.");

		// Return fans to automatic
		if(!InitializeQst())
			exit(EXIT_FAILURE);

		for(i=0;i<MAXFANS;i++)
		{
			if(fan[i].type!=UNUSED)
				SetDutyAutoQst(i);
		}

		// Cleanup
		CleanupQst();

		unlink(HEARTBEAT_FILE);
		exit(EXIT_SUCCESS);
	}
}

int main(int argc,char *argv[])
{
	int val;

	if(argc==2 && !strcmp(argv[1],"--detect"))
	{
		detect();
		exit(EXIT_SUCCESS);
	}

#ifdef DEBUG
	puts("Reading config.");
#endif
	if(readconfig(CONFIG_FILE))
	{
		puts("Invalid or missing config file.");
		exit(EXIT_FAILURE);
	}

	// Initialize Sensor/Controller access
#ifdef DEBUG
	puts("Initializing QST.");
#endif
	if(!InitializeQst())
	{
		UINT32 dwError=(UINT32)errno;
		puts("Cannot initialize QST.");
		return((int)dwError);
	}
	CleanupQst();

	puts("Starting meifand.");

	signal(SIGTERM,handler);
	sigignore(SIGCHLD);

	// Fork to bg
	if(fork()==0)
		fanlogic();

	return(EXIT_SUCCESS);
}
